#include "g_local.h"
#include "BotMain.h"

/***************************************************************************************
original code: TeamWhore

filename: BotAttack.c

Description: This file contains the functions providing help with attacking

###################### Function Descriptions ##################################

qboolean LocateEnemy(gentity_t *bot)
   Locates enemy and judges whether or not the enemy can be seen on distance and visibility

void AttackEnemy(gentity_t *bot, usercmd_t *ucmd, int *state)
	Attacks the enemy by having BUTTON_ATTACK pressed down, sets bot state to ATTACK mode

void FaceEnemy(gentity_t *bot)
	Faces the bots enemy (enemy is defined as bot->enemy)
	Aims at the enemy, taking bot's skill into account

void ChooseWeapon(gentity_t *bot, int preferred)
	Handles the weapon switching when the bot runs out of ammo
	reloading is automatic with cg_autoreload

###############################################################################

-----changes:---------------------
  March 4, 2004 
	rince - added comments to help explain the code
  March 4-5, 2004
    TeamWhore - merged my update/comments with rince's
	    moved HandleAttackState to BotState.c
		added reaction time effect in LocateEnemy
		changed skill-based view distance in LocateEnemy
		more stuff is reset in AttackEnemy
		changed skill-based target acquisition in FaceEnemy
  March 5, 2004
	TeamWhore - changed call to 'visible' to call to 'MyVisible' to make LocateEnemy more accurate
	removed 'state' variable in LocateEnemy definition (unused)
  April 3, 2005
    TeamWhore - moved code in LocateEnemy to improve bots' enemy sighting/selection

****************************************************************************************/

/*///////////////////////////////////
LocateEnemy
loops through every client and checks if the bot can see them
if any are visible, the closest one is selected as the enemy
reaction time upon being shot is based on skill level
/*///////////////////////////////////
qboolean LocateEnemy(gentity_t *bot)
{
	int i;
	gentity_t		*ent;
	gentity_t		*bestenemy = NULL;
	float			bestweight = -1;
	float			weight;
	vec3_t			dist;
	int				attacker;
	int				viewDist;
	int				reactionTime;

	//set the bot's view distance first, so we can check it in the loop
	//and set the reaction time here too, because it's more efficient
	switch(botinfo[bot->s.number].skill)		//different skills see different distances
	{
	case 0:
		viewDist = 1000;		//really short
		reactionTime = 700;		//pretty slow (at least in terms of people, but maybe make this even higher for game)
		break;
	case 1:
		viewDist = 2200;		//short but not usually a problem
		reactionTime = 450;		//maybe slightly slower than most people
		break;
	case 2:
		viewDist = 5000;		//pretty far, will pretty much always see you
		reactionTime = 200;		//faster than almost all people
		break;
	case 3:
		viewDist = 8192;		//REALLY far - I think as far as is ever drawn on screen
		reactionTime = 50;		//one frame - REALLY fast (faster than FRAMETIME, so probably not even significant under 100)
		break;		
	default:					//shut up the compiler
		viewDist = 2200;
		reactionTime = 450;
		break;
	}

	// set the attacker variable to the entity that is shooting the bot and 
	// save entity's client number 
	attacker = bot->client->ps.persistant[PERS_ATTACKER];

	// Make sure the attacker is a valid client from the other team
	if ((attacker != -1) && ((&g_entities[attacker])->client) && (attacker != bot->s.number) && (!OnSameTeam(bot, &g_entities[attacker])))
	{
		int timePassed;		//both are in milliseconds

		//50ms per frame, multiplied by the number of frames passed = time passed
		timePassed = botinfo[bot->s.number].numFramesIdle * 50;
		//if more time has passed than the bot's reaction time (ie, he should react now)
		if (timePassed >= reactionTime)
		{
			// set the bot's 'enemy' to the guy that is hurting the bot
			bot->enemy = &g_entities[attacker];			//if i'm attacked, just go after the attacker
			botinfo[bot->s.number].numFramesIdle = 0;	//reset this because the bot is now reacting
			return qtrue;								//attack!
		}
		botinfo[bot->s.number].numFramesIdle++;			//another frame has passed, increment this so we can keep track of time passed
	}

	//loop through all the clients to see if the bot can see any enemies
	for(i = 0; i < level.maxclients; i++)
	{
		ent = &g_entities[i];

		//make sure the client is a valid player (and in use)
		if (!ent->inuse || !ent->client)
			continue;

		//make sure the player isn't the bot or anyone on his team
		if ((ent == bot) || (ent->client->sess.sessionTeam == bot->client->sess.sessionTeam))
			continue;

		//make sure the player isn't dead or spectating
		if (((ent->health <= 0) || (ent->client->ps.pm_flags & PMF_LIMBO)) || ent->client->sess.sessionTeam == TEAM_SPECTATOR)	//not dead or spectating...
			continue;

		// Check if the bot can see the enemy from this position (can I see him?)
		if (MyVisible(bot, ent) && MyInfront(bot, ent))
		{
			// Find the distance between bot and enemy
			VectorSubtract(bot->r.currentOrigin, ent->r.currentOrigin, dist);

			//store it so we can compare
			weight = VectorLength(dist);

			//if this enemy is too far away to be seen, skip him
			if (weight > viewDist)
				continue;

			//if the new distance is the shortest we have found so far
			if ((weight < bestweight) || (bestweight == -1))
			{
				//set the new shortest distance and the new 'best' enemy
				bestweight = weight;
				bestenemy = ent;	//this loop (and this line) set 'bestenemy' to the closest visible enemy
			}
		}
	}
	
	//if the bot found an enemy
	if(bestenemy)
	{
		bot->enemy = bestenemy;		//set the bot's enemy to 'bestenemy' so the bot can attack him
		return qtrue;				//return yes, he can see an enemy
	}

	return qfalse;		//no, the bot did not find any enemies
}


/*///////////////////////////////////
AttackEnemy
*///////////////////////////////////
void AttackEnemy(gentity_t *bot, int *state)
{
	if (!bot->enemy)									//just a safety check to make sure the bot HAS an enemy
		return;											//return without attacking if he doesn't

	//reset a whole motherload of stuff that shouldn't be set when the bot is attacking
	botinfo[bot->s.number].seekTime = 0;					//bot is attacking a player, so he isn't searching for him
	botinfo[bot->s.number].seekDir = DIR_NOMOVE;			//same as above
	botinfo[bot->s.number].numFramesIdle = 0;				//bot is already attacking, so he's not 'reacting' to anything, so don't count this for reaction time

	FaceEnemy(bot);											//umm...duh?

	ChooseWeapon(bot, 0);									//pick best weapon (currently only based on ammo/charge)
	botinfo[bot->s.number].ucmd->buttons |= BUTTON_ATTACK;	//like pressing 'fire' button 
}

/*//////////////////////////////////////
FaceEnemy
turns the bot to face his enemy (whoever's in bot->enemy)
the lower the bot's skill, the more off he will be
this is essentially how the bots 'aim'
*///////////////////////////////////////
void FaceEnemy(gentity_t *bot)
{
	vec3_t vec, enemyvec;

	// Initialize the enemy's position to the variable
	// (the variable is modified slightly, so don't use bot->enemy.r.currentOrigin)
	VectorCopy(bot->enemy->r.currentOrigin, enemyvec);

	if (bot->enemy->client->ps.pm_flags & PMF_DUCKED)	//the enemy is crouching, so readjust aim
		enemyvec[2] -= bot->enemy->client->ps.crouchViewHeight;

	//TW - this WILL be changed, it is only here to show what can be edited and where
	switch(botinfo[bot->s.number].skill)				//not sure how well this works...basically, target aquisition is affected by skill
	{
	case 0:
		enemyvec[0] += (crandom() * 40);			//crandom is between -1 and 1
		enemyvec[1] += (crandom() * 40);
		enemyvec[2] += (crandom() * 25);			//0 skill is pretty bad aim
		break;
	case 1:
		enemyvec[0] += (crandom() * 30);			//1 skill is better, but not good
		enemyvec[1] += (crandom() * 30);
		enemyvec[2] += (crandom() * 15);
		break;
	case 2:
		enemyvec[0] += (crandom() * 12);			//2 skill is pretty good, but still missed a fair amount of time
		enemyvec[1] += (crandom() * 12);
		enemyvec[2] += (crandom() * 4);
		break;
	case 3:											//3 skill is literally perfect aim...100% of the bot's shots are headshots, and they don't miss
		break;
	}

	//find the direction to the enemy
	VectorSubtract(enemyvec, bot->r.currentOrigin, vec);
	//turn to face the enemy
	vectoangles(vec, bot->s.angles);
}

/*//////////////////////////////
ChooseWeapon
weapon selection for bots
tries to select 'preferred' gun if possible
if impossible or no preferred is specified, sets weapons in decending order
first their normal wepaon (button 3 for people), then their pistol, then a knife
the panz code is supposed to not switch if it's not charged, but in looking over it again
it might be messed up :\
NOTE I don't use the usercmd_t->weapon variable because it doesn't seem to work well with things like grenades, etc

for now, reloading is automatic with cg_autoreload
/*//////////////////////////////
void ChooseWeapon(gentity_t *bot, int preferred)
{
	int i;								//used in 'for' loop
	int weapon;							//the WP_* value for the weapon
	int ammoweap;						//some kind of internal value for the weapon/ammo...don't ask me

	//if a 'preferred' weapon was indicated
	if (preferred != WP_NONE)			//WP_NONE is 0, so passing 0 is no preference
	{
		// Set the ammoweap type to the same type for 'preferred weapon'. 
		// ie: If thompson is preferred weapon, then the ammo type is set to thompson ammo
		ammoweap = BG_FindAmmoForWeapon(preferred);

		// Make sure there is at least 1 round of ammunition for the bot's preferred weapon (in the gun or in the clip)
		if ((bot->client->ps.ammoclip[ammoweap] > 0) || (bot->client->ps.ammo[ammoweap] > 0))
		{
			// We gotta make sure that the panzer is fully charged for use
			if ((preferred != WP_PANZERFAUST) ||												//if it's not a panz
				(level.time - bot->client->ps.classWeaponTime >= g_soldierChargeTime.integer))	//or if the panz is charged
			{
				//we have our preferred weapon, we have ammo for it, and it's ready to go, so switch to it
				bot->client->ps.weapon = preferred;
				//end the function here
				return;
			}
		}
	}

	// Quick reference: MAX_WEAPS_IN_BANK_MP = 8

	for (i = 0; i < MAX_WEAPS_IN_BANK_MP; i++)
	{
		//loop through all the guns in the '3' slot (mp40, thompson, sten, mauser, panzer, venom, flamethrower)
		weapon = weapBanksMultiPlayer[3][i];
		
		// remember we have no ammo, we checked ammo for preferred weapon already from the line: if ((bot->client->ps.ammoclip[ammoweap] > 0) || (bot->client->ps.ammo[ammoweap] > 0))
		if (weapon == preferred)
			continue;		//end the 'for' loop immediately and go to the next value for i

		//if we even have this weapon (since we're looping through them all but only have one of them)
		if (COM_BitCheck(bot->client->ps.weapons, weapon))
		{
			//find the type of ammo for this weapon
			ammoweap = BG_FindAmmoForWeapon(weapon);

			// make sure we have at least 1 round of ammo for the weapon we are checking
			if ((bot->client->ps.ammoclip[ammoweap] > 0) || (bot->client->ps.ammo[ammoweap] > 0))
			{
				//this sets the bot's in-game model to looking through scope, instead of 'running with a mauser'
				if (weapon == WP_MAUSER)		//we don't want them mauser shooting you accurately from a mile away...
					weapon = WP_SNIPERRIFLE;	//not necessary for accuracy, but makes it more like game

				// If the weapon we are checking if the panzerfaust
				if (weapon == WP_PANZERFAUST)
				{
					// Check to make sure it is usable
					if (level.time - bot->client->ps.classWeaponTime < g_soldierChargeTime.integer)		//not charged
						//totally end the 'for' loop - we can only have one of the #3 slot weapons
						//and if we're in this loop, we know it's a panzerfaust (which isn't charged/ready to fire)
						break;
				}

				//if we survived all the checks, then we have ammo, so set the bot's current weapon
				bot->client->ps.weapon = weapon;
				//end the function
				return;
			}
		}
	}
	//the 'break' statement above takes us to here
	
	// Check what team the bot is on
	// Quick reference
	// TEAM_RED   = AXIS
	// TEAM_BLUE  = ALLIED
	if (bot->client->sess.sessionTeam == TEAM_RED)		//if we get here, we don't have any ammo, or panz isn't ready
		weapon = WP_LUGER;	//Since this bot is on the AXIS team, it has a luger 
	else
		weapon = WP_COLT;	//If it's on the ALLIED team, it has a colt

	//again, find the type of ammo for 'weapon'
	ammoweap = BG_FindAmmoForWeapon(weapon);

	//Is there any ammo at all?
	if ((bot->client->ps.ammoclip[ammoweap] > 0) || (bot->client->ps.ammo[ammoweap] > 0))
	{
		// If so we set this as the current weapon
		bot->client->ps.weapon = weapon;
		//end the function here
		return;
	}

	//no ammo for pistols either, so have fun with knives
	// es_fix : maybe we could have the bot commit suicide if the limbo time is nearly zero
	//			but the more realistic fix is to seek out the lieutenant
	//TW - bot now calls for ammo, but not in this func - also doesn't seek lt yet
	bot->client->ps.weapon =  WP_KNIFE;
}
