#ifndef __INCLUDED_BOTMAIN_H_83011__
#define __INCLUDED_BOTMAIN_H_83011__


// es_fix - March 4, 2004: added commments
// TW - March 4-5, 2004: merged rince's comments with mine
//      added several new variables/flags and function definitions
// TW - March 6-7, 2004: added the mod name, new state definitions, a few timeouts,
//			several new function definitions, and variables to support the additions.
//			Moved some of the function definitions/groupings around to be easier to find
//TW - March 9-11, 2004: merged lastMedicChatTime and lastAmmoChatTime (and other chats) into lastChatTime (nod to rince)
//			Added function declaration for new funcs BotChat and Cmd_Push_f (code in g_cmds.c)
//			Removed member bCheered (rendered obsolete by BotChat)

#define		BOT_MOD_NAME				"Ubersoldat"	//the name of this mod :P
#define		UBERSOLDAT_VERSION			2.0f			//stored in the .nod files

// Bot AI states
#define		BOTSTATE_DEAD				-1		//the bot is dead
#define		BOTSTATE_STAND				0		//the bot is standing around doing nothing...this is more just a 'nothing' state
#define		BOTSTATE_MOVE				1		//the bot is moving from one node to another
#define		BOTSTATE_ATTACK				2		//the bot is in combat
#define		BOTSTATE_WANDER				3		//the bot should just wander around for a while
#define		BOTSTATE_FOLLOW				4		//the bot is following a player
#define		BOTSTATE_GIVEITEM			5		//the bot is trying to give someone health/ammo
#define		BOTSTATE_REVIVE				6		//the bot is reviving a wounded player
#define		BOTSTATE_CARRY				7		//the bot has the objective and is trying to bring it to the drop-off point
#define		BOTSTATE_ESCORT				8		//the bot is escorting/protecting another player/bot (who has the obj etc)

// Lets us easily reference movement directions
#define		DIR_NOMOVE					0		//don't move
#define		DIR_FORWARD					1
#define		DIR_RIGHT					2
#define		DIR_BACKWARD				3
#define		DIR_LEFT					4

// Obviously, these are if the bot is ducking/walking/jumping etc
// These appear to be flags from the powers of 2 values (2^0 = 1, 2^1 = 2, 2^2 = 4)
// TW - they are :P  That's how you do flags when more than one can be set at the same time
#define		BOTMOVE_CROUCH				1
#define		BOTMOVE_JUMP				2
#define		BOTMOVE_WALK				4
#define		BOTMOVE_SPRINT				8

//these are timeout values for how long the bot should do certain things
//some of these will probably be changed to be external/user-defined, and/or skill-based etc
#define		SEEK_TIMEOUT				8000	//time a bot should search for an enemy he lost track of
#define		CHAT_REPEATTIME				6000	//time a bot should wait before repeating a chat (medic, ammo, etc)
#define		FOLLOW_TIMEOUT				15000	//time a bot will follow a player who issued a 'follow me' voice command
#define		WANDER_TIMEOUT				15000	//time a bot will wander (not actively seek a long term goal)
#define		SUICIDE_TIMEOUT				12000	//time a bot will wait while stuck somewhere before killing himself


// ************************* These appear to be the node definitions and flags
// Maximum of 1024 nodes for this node set
// TW - node_t nodeset[1024] lets you use 0 - 1023, 1024 is NOT valid :)
#define		MAX_NODES					1024
#define		MAX_LINKS					12

// node flags, so bots know what type of node they are using
#define		NODE_INVALID				-1		// Node is invalid - used so that we (the programmers) know when the bot needs to look for a node
#define		NODE_NORMAL					0		//normal node
#define		NODE_CAMP					1		//camping spot for the bot
#define		NODE_BREAKABLE				2		//node for breakable objects (probably unused because wolfbot does this as a path flag, not a waypoint)
#define		NODE_OBJECTIVE				4		//this node is an objective (see objective flags for type)
#define		NODE_ACTIVATE				8		//this node is something we can activate (door, switch, etc)
#define		NODE_AXIS_UNREACHABLE		16		//node is unreachable for axis
#define		NODE_ALLY_UNREACHABLE		32		//node is unreachable for allies
#define		NODE_MG42					64		//node is at an mg42
#define		NODE_AXIS_DELIVER			128		//place axis should deliver stolen documents/objective
#define		NODE_ALLY_DELIVER			256		//place allies should deliver stolen documents/objective

#define		OBJECTIVE_DYNAMITE			0		//blow this objective up!
#define		OBJECTIVE_STEAL				1		//steal the documents!
#define		OBJECTIVE_CAPTURE			2		//get the flag - not intented for checkpoint, but for spawn flags

// es_fix : added comments to flags
// This way the bot knows how to approach the next node
#define		PATH_NORMAL					0		// Bot moves normally to next node : ie equivalant to pressing the forward move key on your keyboard
#define		PATH_CROUCH					1		//bot should duck and walk toward next node
#define		PATH_SPRINT					2		//bot should sprint to next node
#define		PATH_JUMP					4		//bot should jump to next node
#define		PATH_WALK					8		//bot should walk to next node
#define		PATH_BLOCKED				16		//path to next node is blocked
#define		PATH_LADDER					32		//ladders!

// *******************************************************************************************
// -------------------------------------------------------------------------------------------
// *******************************************************************************************

// Bot info settings
typedef struct botmodinfo_s
{
	qboolean			inuse;						//true if this structure is being used
	short int			state;						//the bot's state
	short int			skill;						//skill of the bot, not only aim - eventually I want to make all this user-configurable
	usercmd_t*			ucmd;						//just to have a copy of the bot's user command, makes life easier on us
	unsigned int		seekTime;					//when the bot started searching for his enemy
	short int			seekDir;					//the direction the bot should move when searching for his enemy
	short int			previousNode;				//name is misleading, node bot was at 2 nodes ago really...useful for certain checks
	short int			currentNode;				//node the bot is at/was just at
	short int			nextNode;					//node the bot is trying to get to
	short int			framesAtNode;				//used to keep track of how long a bot has been at a node, useful for certain node types
	qboolean			movingAtNode;				//will be used in movement code for performing actions at nodes
	unsigned int		lastChatTime;				//last time the bot issued a voice chat
	short int			followingNum;				//entnum of player who issued 'follow me' command
	unsigned int		followingTime;				//time bot started following the player
	short int			giveItemTarget;				//entnum of player who needs ammo/health
	short int			givingSelfItemState;		//used for giving self ammo/health - 0 = not giving self anything, 1 = turned to an open direction, 2 = has thrown item
	short int			numFramesIdle;				//used with reaction time to determine whether or not bot has noticed enemy yet
	short int			reviveTarget;				//entnum of player who needs to be revived
	short int			framesPassed;				//this is different from numFramesIdle because numFramesIdle is used for reaction time calculations
	short int			longTermGoal;				//the number of the node the bot is seeking in the long run
	unsigned int		eventTime[MAX_GENTITIES];	//the time an event occurred
	unsigned int		wanderTime;					//the time a bot started wandering
	short int			pathlist[128];
	short int			pathsize;
	short int			moveFlags;					//a local copy of the movement flags between the bot's current and next nodes
	unsigned int		suicideTime;				//the time at which a bot will commit suicide
	qboolean			onNode;						//true if the bot got as close as he can to a node, helps with BotMoveNearerNode
} botmodinfo_t;

//for the node-specific link/path between 2 nodes
typedef struct nodelink_s
{
	short int	target;		//the node this link is connected to
	short int	flags;		//the path flags for movement along this path
} nodelink_t;

typedef struct node_s
{
	vec3_t		pos;				//the node's position in the world
	short int	num;				//the node's number
	short int	flags;				//what type of node this is (normal, camp, objective, etc)
	short int	objectNum[3];		//id numbers of any world objects associated with this node (used only with unreachable flags)
	int			objFlags;			//objective flags if this node is an objective node
	short int	objTeam;			//the team that should complete this objective
	short int	objType;			//what type of objective this is - see OBJECTIVE_XXX flags
	//short int	objPriority;		//the priority of the objective
	//short int	  objLinkNum;
	short int	objEntity;			//the entity number of what object to dynamite at a dynamite objective node
	short int	numLinks;			//the number of nodes this node is connected to
	nodelink_t	links[MAX_LINKS];
} node_t;


//FUNCTION DECLARATIONS (files are listed in alphabetical order)
// *******************************************************************************************
// -------------------------------------------------------------------------------------------
// *******************************************************************************************

// begin filename : BotAttack.c

qboolean LocateEnemy(gentity_t* bot);
void AttackEnemy(gentity_t* bot, int* state);
void FaceEnemy(gentity_t* bot);
void ChooseWeapon(gentity_t* bot, int preferred);

// end filename : BotAttack.c

// *******************************************************************************************
// -------------------------------------------------------------------------------------------
// *******************************************************************************************

//begin filename : BotMain.c

void Botthink(gentity_t* bot);

//end filename : BotMain.c

// *******************************************************************************************
// -------------------------------------------------------------------------------------------
// *******************************************************************************************

//begin filename : BotMisc.c

void ClearBotInfo(int cn);
void InitBotInfo();
void BotSaveInfo(int clientNum, int num);
void BotLoadInfo();
void UpdateBotInfo();
qboolean AddBot(char* name, char* team, char* skill, int pcl, int weapon, qboolean restart);
void BotParseServerCommand(gentity_t* bot, int* state);
void BotSendCommand(gentity_t* bot, int* state);
void BotCheckAmmoState(gentity_t* bot, int* state);
qboolean BotNeedAmmo(gentity_t* bot, int* state);
void BotGiveItemToPlayer(gentity_t* bot, int* state);
void BotRevivePlayer(gentity_t* bot);
void BotEvents(gentity_t* bot);
void BotHandleEvents(gentity_t* bot, entityState_t* state);
void BotTalk(gentity_t* bot, char* chat, qboolean teamOnly, qboolean checkSpam);

//end filename : BotMisc.c

// *******************************************************************************************
// -------------------------------------------------------------------------------------------
// *******************************************************************************************

//begin filename : BotMove.c

qboolean CanMove(gentity_t* bot, int dir);
void BotMove(gentity_t* bot, int dir, int flags);
qboolean BotMoveNearerNode(gentity_t* bot);
qboolean BotSpecialMove(gentity_t* bot);
void BotFollowPlayer(gentity_t* bot);
int BotFindFreeDirection(gentity_t* bot);

//end filename : BotMove.c

// *******************************************************************************************
// -------------------------------------------------------------------------------------------
// *******************************************************************************************

//begin filename : BotNodes.c

qboolean AddNode(vec3_t origin, int fl, short int* ents, int objFl);
qboolean ConnectNodes(int from, int to, int flags);
void ClearNodes();
void InitNodes();
short int BotFindClosestNode(gentity_t* bot);
short int BotGetNextNode(gentity_t* bot, int* state);
qboolean LoadNodes(char* filename);
void SaveNodes(char* filename);
void BotFindLongTermGoal(gentity_t *bot);
void UpdateNodes();

//end filename : BotNodes.c

// *******************************************************************************************
// -------------------------------------------------------------------------------------------
// *******************************************************************************************

//begin filename : BotState.c

void BotStateActions(gentity_t* bot, int* state);
void HandleDeathState(gentity_t* bot, int* state);
void HandleAttackState(gentity_t* bot, int* state);
void HandleWanderState(gentity_t* bot, int* state);
void HandleFollowState(gentity_t* bot, int* state);
void HandleGiveItemState(gentity_t* bot, int* state);
void HandleReviveState(gentity_t* bot, int* state);
void HandleMoveState(gentity_t* bot, int* state);
void HandleIdleState(gentity_t* bot, int* state);
void HandleCarryObjectiveState(gentity_t* bot, int* state);
qboolean PerformNodeActions(gentity_t* bot);
qboolean BotDoDynamite(gentity_t* bot);
qboolean BotDoStealObjective(gentity_t* bot);
qboolean BotDoFlag(gentity_t* bot);

//end filename : BotState.c

// *******************************************************************************************
// -------------------------------------------------------------------------------------------
// *******************************************************************************************

//begin filename : BotUtils.c

qboolean MyInfront(gentity_t* self, gentity_t* other);
qboolean MyVisible(gentity_t* self, gentity_t* other);
int GetCost(int from, int to, int hops);
int CreatePathAStar(gentity_t *bot, int from, int to, short int *pathlist);
int GetFCost(int to, int num, int parentNum, float *gcost);

//end filename : BotUtils.c

// *******************************************************************************************
// -------------------------------------------------------------------------------------------
// *******************************************************************************************

#endif

