#include "g_local.h"
#include "BotMain.h"


/***************************************************************************************
original code: TeamWhore

filename: BotMain.c

Description: This file contains the Bot think function 'Botthink'

###################### Function Descriptions ##################################

void ClearBotInfo(int cn)
  Clears all the members of the botinfo structure for the specified client number

void InitBotInfo()
  Clears the botinfo structure for all client numbers

void BotSaveInfo(int clientNum, int num)
  Saves information about the bot in a global variable (cvar I think)

void BotLoadInfo()
  Loads bot information that was saved

void UpdateBotInfo()
  Checks whether or not any bots have disconnected, and clears the corresponding botinfo
	structures of those who have

qboolean AddBot(char* name, char* team, char* skill, int pcl, int wpn, qboolean restart)
    Adds bot based on parameters

void BotParseServerCommand(gentity_t *bot, int *state)
	Prevent bot from timing out, also allows us to handle server events like chat messages

void BotSendCommand(gentity_t *bot, int *state)
	Performs final actions per frame and sends bot's command to server

void BotCheckAmmoState(gentity_t *bot, int *state)
	Checks if the bot needs ammo and handles the bot's response to this situation

qboolean BotNeedAmmo(gentity_t *bot, int *state)
	Checks if the bot needs ammo - called by BotCheckAmmoState

void BotGiveItemToPlayer(gentity_t *bot)
	Handles when the bot is trying to give a player health/ammo

void BotRevivePlayer (gentity_t *bot)
	Handles when a bot is trying to revive a wounded player

void BotEvents (gentity_t *bot)
	Loops through all the entities and passes them to the bot so that we can parse events

void BotHandleEvents (gentity_t *bot, entityState_t *state)
	Handles the parsing of events (sounds, certain things happening...)

void BotTalk(gentity_t *bot, char *chat, qboolean teamOnly, qboolean checkSpam)
	A wrapper function to implement voice chats for bots, taking into account spam prevention

###############################################################################

-----changes:---------------------
  March 4, 2004 
	rince - added comments to help explain the code
  March 4-5, 2004
    TeamWhore - merged rince's comments with mine
	   added default values for new variables in InitBotInfo
	   added team chat handling for 'follow me' and 'i need ammo' voice chats in BotParseServerCommand
	   added function BotCheckAmmoState to handle if the bot needs ammo
	   added function BotNeedAmmo to check if the bot needs ammo
	   added function BotGiveItemToPlayer to handle if the bot is trying to give a player ammo/health
	   added function BotFindFreeDirection to search for an open direction in the immediate area around the bot
  March 5, 2004
	TeamWhore - added check inside BotCheckAmmoState to only call ChooseWeapon if the bot previously changed guns to give himself ammo
		   this fixes a few bugs relating to bots shooting instead of giving out ammo/health etc
	   fixed bug where lt giving himself ammo wouldn't throw ammo bag but just run forward - only occurred with thompson
	   moved MyInfront here from BotMove.c
	   added function MyVisible to allow bots to sight enemies better
	   moved BotFindFreeDirection to BotMove.c
	   added handling for 'medic' voice chat
	   added function BotRevivePlayer
	   edited BotGiveItemToPlayer to make the bot wait a few frames when giving the item - this fixes bugs with the movement where the
		  bot would throw the bag as he turned around, and give it to himself instead of the player
  March 6-7, 2004
	TeamWhore - added BotSendCommand function to handle final actions each frame
	   fixed the 'dotConst' values in MyInfront to correctly represent what I said they did :P
	   added BotEvents function to loop through all entities and pass them to the bot's handler
	   added BotHandleEvents function to catch/respond to any events we want to process
	   Bots now sometimes taunt a member of the opposing team if he kills himself or falls to his death
	   Bots now sometimes cheer/say 'disarm the dynamite' after a team plants dynamite at an objective
  March 11, 2004
	TeamWhore - added function BotTalk to clean up bot voice-chat code and to internalize all the spamming checks
	   Modified sections of code that used bot chats to incorporate this
	   Removed bCheered variable and all it's references
	   Added ClearBotInfo and UpdateBotInfo, and changed InitBotInfo to a looping call to ClearBotInfo
  March 13, 2004
	TeamWhore - removed all references to 'think' and 'nextthink' for bots (BotSendCommand, maybe others)
  April 3, 2004
	TeamWhore - moved a lot of functions into new file BotUtils.c
	   Added initialization for new variables in botinfo structure
	   Fixed a typo in one of the comments - useful!


****************************************************************************************/

/*////////////////////////////////////
ClearBotInfo
'cn' is the client number of the info to clear
Called to clear the botinfo structure for a specific
client number - called at every server-load and when a
bot disconnects
*/////////////////////////////////////
void ClearBotInfo(int cn)
{
	botinfo[cn].inuse = qfalse;							//there are no bots, so this structure is unused
	botinfo[cn].state = BOTSTATE_DEAD;					//default state for bots - so we know to initialize more of their info
	botinfo[cn].skill = 0;								//this will get set when the bot is added in AddBot
	botinfo[cn].ucmd = NULL;							//clear the usercommand because it doesn't exist yet
	botinfo[cn].previousNode = NODE_INVALID;
	botinfo[cn].currentNode = NODE_INVALID;
	botinfo[cn].nextNode = NODE_INVALID;				//set all the nodes to invalid because the bot doesn't exist yet
	botinfo[cn].movingAtNode = qfalse;					//a non-existant bot isn't moving :P
	botinfo[cn].framesAtNode = 0;						//and hasn't spent any time at a node
	botinfo[cn].seekDir = DIR_NOMOVE;					//still not moving, or seeking an enemy
	botinfo[cn].seekTime = 0;							//ditto above
	botinfo[cn].followingNum = ENTITYNUM_NONE;			//he isn't following anyone
	botinfo[cn].followingTime = 0;						//ditto above
	botinfo[cn].giveItemTarget = ENTITYNUM_NONE;		//he's not giving anyone ammo/health
	botinfo[cn].givingSelfItemState = 0;				//ditto above
	botinfo[cn].lastChatTime = 0;						//hasn't talked yet
	botinfo[cn].numFramesIdle = 0;						//and he doesn't have a reaction time because he doesn't exist
	botinfo[cn].reviveTarget = ENTITYNUM_NONE;			//he isn't trying to revive anyone
	botinfo[cn].framesPassed = 0;						//obviously no frames have passed
	botinfo[cn].longTermGoal = NODE_INVALID;			//check anew for the goal
	botinfo[cn].wanderTime = 0;							//bot isn't wandering
	botinfo[cn].pathsize = 0;							//doesn't have a goal, so no path
	botinfo[cn].moveFlags = PATH_NORMAL;				//bot isn't moving anywhere yet
	botinfo[cn].suicideTime = 0;
	botinfo[cn].onNode = qfalse;

	memset(botinfo[cn].pathlist, NODE_INVALID, 128);
	memset(botinfo[cn].eventTime, 0, MAX_GENTITIES);
}


/*////////////////////////////////////
InitBotInfo

called in G_InitGame
Just calls ClearBotInfo for every structure
*/////////////////////////////////////
void InitBotInfo()
{
	int i;

	for (i = 0; i < MAX_CLIENTS; i++)
		ClearBotInfo(i);
}


/*/////////////////////////////////////
BotSaveInfo

writes the necessary bot info (name, team, class, weapon and skill) to a cvar
so we can reload it after map change/restart
*//////////////////////////////////////
void BotSaveInfo(int clientNum, int num)
{
	const char	*s;
	const char	*var;
	gclient_t *client = level.clients + clientNum;
	char userinfo[MAX_INFO_STRING];
	char skill[MAX_STRING_CHARS], name[MAX_STRING_CHARS];
	int team, wpn, pclass;
	char *buf;

	//get the userinfo for the bot
	trap_GetUserinfo(clientNum, userinfo, MAX_INFO_STRING);

	//save the skill of the bot
	buf = Info_ValueForKey(userinfo, "skill");
	strcpy(skill, buf);

	//save the bot's team
	team = client->sess.sessionTeam;
	//save the bot's player class (soldier, med, eng, lt)
	pclass = client->sess.playerType;
	//save the bot's current main weapon
	wpn = client->sess.playerWeapon;
	//save the bot's name
	strcpy(name, client->pers.netname);

	//this line puts all the saved info into one string
	s = va("%s %i %i %i %s", name, team, pclass, wpn, skill);

	//this line creates a string "botinfo#", where # is which number bot this is
	//so if we have 5 bots, there will be botinfo0 to botinfo4 - (0,1,2,3,4) is 5 numbers
	var = va("botinfo%i", num);

	//actually save the data - creates a cvar named "botinfo#" with the value of the string s
	trap_Cvar_Set(var, s);
}


/*////////////////////////////////////
BotLoadInfo

called after map change/restart to reload all the botinfo we saved
then re-adds all the bots with this info, so it appears that they
never left
*/////////////////////////////////////
void BotLoadInfo()
{
	int i;
	const char *var;
	char s[MAX_STRING_CHARS];
	char name[MAX_STRING_CHARS], skill[MAX_STRING_CHARS];
	int team, pclass, wpn;
	char tm[5];

	//so we add all the bots that were used last round
	for (i = 0; i < numBots; i++)
	{
		//read in the data we saved in BotSaveInfo
		var = va("botinfo%i", i);
		trap_Cvar_VariableStringBuffer(var, s, sizeof(s));

		//extract out all the information - name, team, player class, main weapon and skill
		sscanf(s, "%s %i %i %i %s", &name, &team, &pclass, &wpn, &skill);

		//make sure the bot is on the correct team
		if (team == TEAM_RED)
			strcpy(tm, "red");
		else
			strcpy(tm, "blue");

		//add the bot back into the game
		if (!AddBot(name, tm, skill, pclass, wpn, qtrue))
			G_Printf(S_COLOR_YELLOW, "Error adding bot %s\n", name);
	}
}

/*////////////////////////////////////
UpdateBotInfo
This function is called every frame to reset
the botinfo structures of any bots who disconnected
(kicked etc).  This is necessary so that old data doesn't
get transferred to a new bot, and so that numBots will be correct
for level changes etc.
*/////////////////////////////////////
void UpdateBotInfo()
{
	int i;
	for (i = 0; i < MAX_CLIENTS; i++)
	{
		gentity_t *ent = &g_entities[i];

		if ((botinfo[i].inuse) && ((!ent->client) || (ent->client->pers.connected != CON_CONNECTED)))
		{
			ClearBotInfo(i);
			numBots--;			//so that it will be correct (we hope)
		}
	}
}

/*////////////////////////////////////
AddBot

actually adds the bot into the game and sets all the info
then places the bot on the correct team
'restart' is passed in so that the message "<name> has joined the <team> team"
won't be displayed when the bots are re-added after a map change/reload
Used in Svcmd_Addbot_f() and BotLoadInfo
/*////////////////////////////////////
extern void Botthink(gentity_t* bot);

qboolean AddBot(char* name, char* team, char* skill, int pcl, int wpn, qboolean restart)
{
	int				clientNum;
	gentity_t		*bot;
	char			userinfo[MAX_INFO_STRING];

	//check if there are any available slots for the bot
	clientNum = trap_BotAllocateClient();
	if (clientNum == -1)
	{
		G_Printf("%sUnable to add bot, all client spaces full\n", S_COLOR_RED);
		return qfalse;
	}

	userinfo[0] = '\0';

	//set the client variables for the bot
	Info_SetValueForKey(userinfo, "name", name);
	Info_SetValueForKey(userinfo, "rate", "25000");
	Info_SetValueForKey(userinfo, "snaps", "20");
	Info_SetValueForKey(userinfo, "skill", skill);
	Info_SetValueForKey(userinfo, "team", team);
	Info_SetValueForKey(userinfo, "cg_autoactivate", "1");	//PICKUP_TOUCH - lets bots pickup health/ammo etc
	Info_SetValueForKey(userinfo, "cg_autoreload", "1");	//so we don't have to handle it - maybe change later

	trap_SetUserinfo(clientNum, userinfo);

	//check if the bot can connect - do NOT pass qtrue for the 3rd paramater (isBot)!
	if (ClientConnect(clientNum, qtrue, qfalse))			//only returns a value if it failed
		return qfalse;

	//set some generic info
	bot									= &g_entities[clientNum];
	bot->r.svFlags						|= SVF_BOT;			//this is how we know he's a bot :P
	bot->inuse							= qtrue;
	bot->scriptStatus.scriptEventIndex	= -1;
	bot->s.number						= bot - g_entities;
	bot->r.ownerNum						= ENTITYNUM_NONE;
	bot->headshotDamageScale			= 1.0;
	bot->client->pers.localClient		= qtrue;			//not really necessary, but true...means it's on the server computer
															//this prevents anyone from filing complaints against tk'ing bots

	botinfo[bot->s.number].inuse		= qtrue;			//so we know this structure is being used
	botinfo[bot->s.number].skill		= atoi(skill);		//set the bot's skill level

	//set the bot's player class
	bot->client->ps.stats[STAT_PLAYER_CLASS]	= pcl;
	bot->client->sess.latchPlayerType			= pcl;		//'latch' values are what you'd set in the limbo menu
	bot->client->sess.playerType				= pcl;

	//set the bot's main weapon
	bot->client->sess.latchPlayerWeapon			= wpn;
	bot->client->sess.playerWeapon				= wpn;

	SetTeam(bot, team, restart);							//put him on the team and spawn him

	if (!restart && g_botAllowRadio.integer && random() < 0.65)		//make some of them say hi but not always
		BotTalk(bot, "hi", qfalse, qfalse);	

	if (!restart)
		numBots++;

	return qtrue;
}

/*////////////////////////////////////////
BotParseServerCommand
Captures the server commands and pretends to handle them
so the bots don't get 'server command overflow' errors
Also allows us to parse server messages like chats
*/////////////////////////////////////////
void BotParseServerCommand(gentity_t *bot, int *state)
{
	char buf[1024];
	char *args;

	while (trap_BotGetServerCommand(bot->client->ps.clientNum, buf, sizeof(buf)))
	{
		args = strchr(buf, ' ');
		if (!args) 
			continue;
		*args++ = '\0';

		Q_CleanStr(args);

		if (!Q_stricmp(buf, "cp"))
		{ /*CenterPrint*/ }
		else if (!Q_stricmp(buf, "cs"))
		{ /*ConfigStringModified*/ }
		else if (!Q_stricmp(buf, "print"))
		{ /*CenterPrintf*/ }
		else if (!Q_stricmp(buf, "chat"))
		{ /*CenterPrintf*/ }
		else if (!Q_stricmp(buf, "tchat"))
		{ /*CenterPrintf*/ }
		else if (!Q_stricmp(buf, "vtchat"))	//this is a team voice chat
		{
			qboolean voiceOnly;	//we don't use this, but it's sent so we have to read it - whether or not the chat prints text also
			int clientNum;		//client number of the person saying the chat
			char color;			//the color to print the text in - we don't use this either
			char chat[20];		//chat is the chat name, like 'needammo' or 'defendobjective'

			if (bot->client->ps.pm_type == PM_DEAD)		//don't try to follow orders when dead
				return;

			//read in the data that was passed by the server
			sscanf(args, "%d %d %d %s", &voiceOnly, &clientNum, &color, chat);

			if (!Q_stricmp(chat, "followme"))	//if someone said 'follow me'
			{
				//if he's not already following someone, or trying to give anyone health/ammo or revive anyone
				if ((*state != BOTSTATE_ATTACK) && (*state != BOTSTATE_FOLLOW) && (*state != BOTSTATE_GIVEITEM) && (*state != BOTSTATE_REVIVE))
				{
					vec3_t v;
					float dist;
					gentity_t* player = &g_entities[clientNum];		//the player who said the chat

					if (player->s.number == bot->s.number)
						continue;									//don't follow yourself

					//find where the player is and the distance to him
					VectorSubtract(player->r.currentOrigin, bot->r.currentOrigin, v);
					dist = VectorLength(v);

					if (dist < 600)		//if he's relatively close (so we know he was talking to this bot)
					{
						//if the bot can talk, acknowledge the command by saying 'affirmative'
						BotTalk(bot, "affirmative", qtrue, qtrue);
						//tell the bot to follow this player
						botinfo[bot->s.number].followingNum = clientNum;
						//set the time the bot started following this player
						botinfo[bot->s.number].followingTime = level.time;
						*state = BOTSTATE_FOLLOW;	//let the bot know to start following this player
					}
				}
			}
			else if (!Q_stricmp(chat, "needammo"))		//if someone said 'i need ammo'
			{
				//if the bot is a lt and not already trying to give someone ammo
				if ((bot->client->ps.stats[STAT_PLAYER_CLASS] == PC_LT) && (*state != BOTSTATE_GIVEITEM) && (*state != BOTSTATE_ATTACK))
				{
					//if the bot has enough charge in his bar (ammo packs take 1/4 bar)
					if ((level.time - bot->client->ps.classWeaponTime) >= g_LTChargeTime.integer*0.25f)
					{
						vec3_t v;
						float dist;
						gentity_t* player = &g_entities[clientNum];	//the player who said the chat

						if (player->s.number == bot->s.number)
							continue;								//don't be talkin' to yo'self no more boy!

						//figure out where the player is and how far away
						VectorSubtract(player->r.currentOrigin, bot->r.currentOrigin, v);
						dist = VectorLength(v);

						//if he's relatively close (so bot's don't run across level trying to give ammo)
						if (dist < 800)
						{
							//if the bot can talk, acknowledge command
							BotTalk(bot, "affirmative", qtrue, qtrue);

							botinfo[bot->s.number].giveItemTarget = clientNum;		//set up guy to get ammo
							*state = BOTSTATE_GIVEITEM;
						}
					}
				}
			}
			else if (!Q_stricmp(chat, "medic"))		//if someone called for a medic
			{
				//if the bot is a medic and he's not trying to revive someone already
				if ((bot->client->ps.stats[STAT_PLAYER_CLASS] == PC_MEDIC) && (*state != BOTSTATE_REVIVE))
				{
					gentity_t *player = &g_entities[clientNum];		//the player who said the chat
					vec3_t v;
					float dist;

					if (player->s.number == bot->s.number)	//don't react to your own chats
						continue;

					//figure out where the player is and how far away
					VectorSubtract(player->r.currentOrigin, bot->r.currentOrigin, v);
					dist = VectorLength(v);

					if (player->client->ps.pm_type == PM_DEAD)	//if the player is dead and needs a revive
					{
						if (dist < 1100)	//make bots move a good distance to revive people
						{
							//if bot can talk and dead player isn't another bot, let player know he's coming
							if (!(player->r.svFlags & SVF_BOT))
								BotTalk(bot, "iammedic", qtrue, qtrue);

							*state = BOTSTATE_REVIVE;
							botinfo[bot->s.number].reviveTarget = clientNum;	//set up the guy to revive
							ChooseWeapon(bot, WP_MEDIC_SYRINGE);		//select the syringe
						}
					}
					else if (*state != BOTSTATE_GIVEITEM)		//the player just wants a health bag, don't give him one if i'm busy
					{
						//if the bot has enough charge in his charge bar (medpacks are 1/4 of the bar)
						if ((level.time - bot->client->ps.classWeaponTime) >= g_medicChargeTime.integer*0.25f)
						{
							if (dist < 800)		//if the player is relatively close
							{
								BotTalk(bot, "affirmative", qtrue, qtrue);

								botinfo[bot->s.number].giveItemTarget = clientNum;	//set up the guy to get health
								*state = BOTSTATE_GIVEITEM;
							}
						} //charged
					} //player dead
				} //medic
			} //'medic' chat
		}
		else if (!Q_stricmp(buf, "scores"))
		{ /*FIXME: parse scores?*/ }
		else if (!Q_stricmp(buf, "clientLevelShot"))
		{ /*ignore*/ }
	}
}

//this function would normally get called by all clients anyway, here I use it explicitly
extern void ClientThink_real(gentity_t* ent);
/*////////////////////////////////////////////////////
BotSendCommand
This function just sets some final values after we're done parsing
all the data each frame, and sends it off to the server
*/////////////////////////////////////////////////////
void BotSendCommand(gentity_t *bot, int *state)
{
	botinfo[bot->s.number].state = *state;
	botinfo[bot->s.number].ucmd->serverTime = level.time;			//gets rid of phonejack icon for ConnInts

	//ANGLE2SHORT just converts the the angles to a value from 0 to 65535 (used internally, I guess)
	botinfo[bot->s.number].ucmd->angles[PITCH] = ANGLE2SHORT(bot->s.angles[PITCH]);
	botinfo[bot->s.number].ucmd->angles[YAW] = ANGLE2SHORT(bot->s.angles[YAW]);
	botinfo[bot->s.number].ucmd->angles[ROLL] = ANGLE2SHORT(bot->s.angles[ROLL]);	//execute turning (set the bot's angles)

	bot->client->pers.cmd = *botinfo[bot->s.number].ucmd;		//store the bot's user command

	trap_BotUserCommand(bot->client->ps.clientNum, botinfo[bot->s.number].ucmd);	//send the bot's command
	ClientThink_real(bot);
}

/*///////////////////////////////////////////////////////
BotCheckAmmoState
This function checks if the bot needs ammo and handles it
If he's a lt, he gives himself ammo, otherwise he'll call for ammo
The handling of giving himself ammo has a few issues, but is usually ok
*////////////////////////////////////////////////////////
void BotCheckAmmoState(gentity_t *bot, int *state)
{
	if (BotNeedAmmo(bot, state))		//if the bot needs ammo
	{
		//if the bot is a lt, give himself ammo rather than ask for it
		if (bot->client->ps.stats[STAT_PLAYER_CLASS] == PC_LT)
		{
			botinfo[bot->s.number].giveItemTarget = bot->s.number;
			//don't set state here, because we allow bot to give himself ammo in BOTSTATE_ATTACK
		}
		//if the bot is not a lt, ask for ammo
		else
		{
			BotTalk(bot, "needammo", qtrue, qtrue);
		}
	}
	//if he doesn't need ammo
	else
	{
		//only do this if he previously did need ammo (so he had the ammo bag out)
		//otherwise there can be issues with the bots shooting when they should be throwing items/dyno, etc
		if (botinfo[bot->s.number].givingSelfItemState != 0)
		{
			botinfo[bot->s.number].givingSelfItemState = 0;		//(re)set this here, cuz now we know he picked up/got ammo
			*state = BOTSTATE_STAND;
			ChooseWeapon(bot, 0);								//select the gun again
		}
	}
}


/*///////////////////////////////////////////////////////
BotNeedAmmo
This function just checks if the bot needs ammo for his main gun
and for whatever weapon he's currently using, if different (ie, pistol)
Currently, 'needing ammo' for a weapon is defined as not having any ammo for the main gun
or (if using pistol - cuz panz is charging etc), not having any extra clips
*////////////////////////////////////////////////////////
qboolean BotNeedAmmo(gentity_t *bot, int* state)
{
	int i;
	int ammoweap, weapon;
	qboolean mainAmmo = qfalse;

	for (i = 0; i < MAX_WEAPS_IN_BANK_MP; i++)
	{
		weapon = weapBanksMultiPlayer[3][i];		//loop through all the main weapons (like hitting '3' in game)

		if (COM_BitCheck(bot->client->ps.weapons, weapon))		//do I have this weapon
		{
			//get the type of ammo for this weapon
			ammoweap = BG_FindAmmoForWeapon(weapon);

			//do I have any ammo in this gun, or clips for this gun
			if ((bot->client->ps.ammo[ammoweap] > 0) || (bot->client->ps.ammoclip[ammoweap] > 0))
				mainAmmo = qtrue;

			//leave this 'for' loop
			break;
		}
	}

	if (!mainAmmo)								//if I don't have any ammo for main weapon, I need ammo
		return qtrue;

	if (bot->client->ps.weapon == weapon)		//if I have ammo for the main gun and I'm using the main gun, I don't need ammo
		return qfalse;

	//we will only get this far if the bot is a panzer, and has panz ammo, but is using a pistol cuz his panz isn't charged

	//sets 'weapon' to whichever pistol the bot's team uses...
	//equivalent to "if the bot's team is red(axis), set his weapon to the luger, else set it to the colt"
	weapon = (bot->client->sess.sessionTeam == TEAM_RED) ? WP_LUGER : WP_COLT;
	//find the ammo type for this weapon
	ammoweap = BG_FindAmmoForWeapon(weapon);

	if (bot->client->ps.ammoclip[ammoweap] > 0)		//if i'm using a pistol and have a spare clip for it (and have ammo for the main gun)
		return qfalse;								//then i don't need ammo

	return qtrue;			//if i get this far, i need ammo
}


/*////////////////////////////////////////////////////////////////////
BotGiveItemToPlayer
This function handles when a bot is trying to give someone health/ammo
Select the health/ammo as the weapon BEFORE calling this
*////////////////////////////////////////////////////////////////////
void BotGiveItemToPlayer(gentity_t *bot, int *state)
{
	vec3_t v;
	float dist;
	gentity_t *player;


	//let a few frames pass, so that we ensure the bot has given the pack - otherwise he turns before he's done throwing it, and gives himself the item instead
	if ((botinfo[bot->s.number].framesPassed > 0) && (botinfo[bot->s.number].framesPassed < 4))
	{
		botinfo[bot->s.number].framesPassed++;		//another frame has passed
		return;				//end the function because we don't want to do anything else
	}
	
	if (botinfo[bot->s.number].framesPassed == 4)	//by now, he should have given the ammo
	{
		botinfo[bot->s.number].giveItemTarget = ENTITYNUM_NONE;		//reset this now
		botinfo[bot->s.number].framesPassed = 0;					//reset this too since we're done
		*state = BOTSTATE_STAND;
		return;		//end the function
	}

	player = &g_entities[botinfo[bot->s.number].giveItemTarget];	//the player the bot wants to give ammo/health to

	//if they're dead or gone, stop trying to give them stuff
	if ((player->client->ps.pm_type == PM_DEAD) || (player->client->pers.connected == CON_DISCONNECTED))
	{
		//reset the item target to nobody
		botinfo[bot->s.number].giveItemTarget = ENTITYNUM_NONE;
		botinfo[bot->s.number].framesPassed = 0;
		*state = BOTSTATE_STAND;
		return;		//end this function
	}

	//find out where the player is and how far away he is
	VectorSubtract(player->r.currentOrigin, bot->r.currentOrigin, v);
	dist = VectorLength(v);
	vectoangles(v, bot->s.angles);			//look at player

	if (dist < 120)				//if he's close enough
	{
		botinfo[bot->s.number].ucmd->buttons |= BUTTON_ATTACK;		//throw an ammo/health bag
		botinfo[bot->s.number].framesPassed = 1;					//let us know that he's thrown it (supposedly)
	}
	else
		BotMove(bot, DIR_FORWARD, 0);	//if he's too far away, move toward him
}

/*///////////////////////////////////////////////
BotRevivePlayer
Handles when the bot is attempting to revive a wounded player
Select the syringe before you call this
*////////////////////////////////////////////////
void BotRevivePlayer(gentity_t *bot)
{
	vec3_t v;
	float dist;
	gentity_t *player = &g_entities[botinfo[bot->s.number].reviveTarget];	//the player the bot wants to revive

	//if they're alive or disconnected, don't try to revive them
	if ((player->client->ps.pm_type != PM_DEAD) || (player->client->pers.connected == CON_DISCONNECTED))
	{
		//reset the revive target to nobody
		botinfo[bot->s.number].reviveTarget = ENTITYNUM_NONE;
		ChooseWeapon(bot, 0);		//select gun again
		return;		//end this function
	}

	VectorCopy(player->r.currentOrigin, v);		//copy the player's position
	v[2] += player->client->ps.mins[2];			//basically, look at the ground under the player (I hope)

	VectorSubtract(v, bot->r.currentOrigin, v);		//figure out the direction to the player
	dist = VectorLength(v);							//and the distance
	vectoangles(v, bot->s.angles);					//look at player

	if (dist < 48)				//if he's close enough - taken from Weapon_Syringe
	{
		BotMove(bot, DIR_NOMOVE, BOTMOVE_CROUCH);		//stay crouching for the revive
		botinfo[bot->s.number].ucmd->buttons |= BUTTON_ATTACK;		//use syringe
	}
	else if (dist < 100)	//if he's pretty close, start to crouch
		BotMove(bot, DIR_FORWARD, BOTMOVE_CROUCH);
	else
		BotMove(bot, DIR_FORWARD, 0);	//if he's too far away, move toward him
}

/*////////////////////////////////////////////////////
BotEvents
This function just loops through all the entities and calls
BotHandleEvents for each
*/////////////////////////////////////////////////////
void BotEvents(gentity_t *bot)
{
	int i;
	for (i = 0; i < level.num_entities; i++)
		BotHandleEvents(bot, &(g_entities[i].s));
}


/*///////////////////////////////////////////////////
BotHandleEvents
This function checks for any events that have occurred so that
we can parse/handle them as we want.  Examples of events we might
handle are sounds (objective taken, footsteps, etc), players dying
in certain ways, etc
*////////////////////////////////////////////////////
void BotHandleEvents(gentity_t *bot, entityState_t *state)
{
	int event;

	//if we've already handled this event, return (they typically get sent 3 times)
	if (botinfo[bot->s.number].eventTime[state->number] == g_entities[state->number].eventTime)
		return;

	//basically mark that we have already handled this event
	botinfo[bot->s.number].eventTime[state->number] = g_entities[state->number].eventTime;

	//convert the event type, this code taken from Id's bot code
	if (state->eType > ET_EVENTS)
		event = (state->eType - ET_EVENTS) & ~EV_EVENT_BITS;
	else
		event = state->event & ~EV_EVENT_BITS;

	switch (event)										//handle whatever events we want to
	{
	case EV_OBITUARY:									//someone died
		{
			int attacker, target, mod;

			target = state->otherEntityNum;
			attacker = state->otherEntityNum2;
			mod = state->eventParm;

			//if the person who died is on the other team, and died by falling or killing himself
			if (((mod == MOD_FALLING) || (attacker == target)) && (target != bot->s.number) && (!OnSameTeam(bot, &g_entities[target])))
			{
				if (random() < 0.1)		//random is low so they don't all talk
					BotTalk(bot, "greatshot", qfalse, qtrue);
			}

			break;
		}
	case EV_GLOBAL_SOUND:							//a sound is played to everyone
		{
			char buf[MAX_STRING_CHARS];

			trap_GetConfigstring(CS_SOUNDS + state->eventParm, buf, sizeof(buf));		//get the sound

			if (!Q_stricmp(buf, "sound/multiplayer/allies/a-dynamite_planted.wav"))		//allies planted dyno
			{
				if (bot->client->sess.sessionTeam == TEAM_BLUE)				//if this bot is on allies
				{
					if (random() < 0.1)
						BotTalk(bot, "cheer", qfalse, qtrue);		//sometimes cheer
				}
				else
				{
					if (random() < 0.05)
						BotTalk(bot, "disarmdynamite", qtrue, qtrue);		//axis sometimes say 'disarm the dynamite'
				}
			}
			else if (!Q_stricmp(buf, "sound/multiplayer/axis/g-dynamite_planted.wav"))	//axis planted dyno
			{
				if (bot->client->sess.sessionTeam == TEAM_RED)				//if this bot is on axis
				{
					if (random() < 0.1)
						BotTalk(bot, "cheer", qfalse, qtrue);		//sometimes cheer
				}
				else
				{
					if (random() < 0.05)
						BotTalk(bot, "disarmdynamite", qtrue, qtrue);		//allies sometimes say 'defuse the dynamite'
				}
			}
			break;
		}
	}
}

/*////////////////////////////////////////////
BotTalk
This function just provides a simpler way for us to
make the bots talk, but not spam chats.
This way we don't need time/spam checks every time they talk
If checkSpam is qfalse, the chat will be played regardless of the
time since the last chat
*/////////////////////////////////////////////
extern void G_Voice(gentity_t *ent, gentity_t *target, int mode, const char *id, qboolean voiceonly);

void BotTalk(gentity_t *bot, char *chat, qboolean teamOnly, qboolean checkSpam)
{
	int mode;

	if (!g_botAllowRadio.integer)	//return if they're not allowed to talk
		return;

	//if not enough time has passed, and this is time-dependant
	if ((checkSpam) && (level.time < (botinfo[bot->s.number].lastChatTime + CHAT_REPEATTIME)))
		return;

	mode = teamOnly ? SAY_TEAM : SAY_ALL;	//mode is SAY_TEAM if teamOnly is true, otherwise it's SAY_ALL

	G_Voice(bot, NULL, mode, chat, qfalse);	//actually do the chat

	if (checkSpam)
		botinfo[bot->s.number].lastChatTime = level.time;	//mark the chat time for spam checking
}
