#include "g_local.h"
#include "us_misc.h"

/***************************************************************************************
original code: TeamWhore

filename: us_cmds.c

Description: This file contains functions implementing and related to various client commands

###################### Function Descriptions ##################################

void Cmd_Gib_f(gentity_t *ent)
	Implements instant self-gibbing via the /gib command

void Cmd_Clientnum_f(gentity_t *ent)
	Gives players the ability to retrieve the client numbers of other players on the server

void Cmd_Stats_f(gentity_t *ent)
	Implements individual statistics for players, displaying various information about their
		play for each round

void Cmd_Omw_f(gentity_t *ent)
	Allows players to use the "Affirmative, on my way" sound as a team-only voice chat

void Cmd_Sticky_f(gentity_t *ent)
	Allows players to toggle between normal and sticky grenades with the /sticky command

void Cmd_Smoke_f(gentity_t *ent)
	Allows LTs to toggle between normal airstrike canisters and smoke grenades/canisters

void Cmd_Playdead_f(gentity_t *ent)
	Allows players to fake death (and to return to normal) with the /playdead command

void Cmd_Pmsg_f(gentity_t *ent)
	Allows players to send private text-only messages to one or more other players

void Cmd_Push_f(gentity_t *ent)
	Implements shrubmod-like shoving with the +salute command

void Cmd_ThrowKnife_f(gentity_t *ent)
	Implements knife throwing with the +dropweapon command (usually the 'p' key)

void Touch_Knife(gentity_t *ent, gentity_t *other, trace_t *trace)
	Handler function for when a client touches a knife that was thrown

void Cmd_KickGrenade_f(gentity_t *ent)
	Implements grenade kicking with the +salute command, if the player is near a grenade

void Cmd_KickAS_f(gentity_t *ent)
	Allows a LT to kick his own airstrike canisters with the +salute command

void weapon_smokeGrenade(gentity_t *ent)	NOTE - will likely move to a different file in the future
	The 'think' function for smoke grenades, allows them to smoke and emit a hissing sound

qboolean PlayerNeedsAmmo(gentity_t *ent)	NOTE - will likely move to a different file in the future
	Checks whether a player currrently has any use for an ammo pack.  If not, we
		can prevent them from picking up the pack and whoring ammo

int ClientNumberFromNameMatch(char *name, int *matches)		NOTE - will likely move to a different file
	Takes a string representing a partial name, and returns the client numbers of 
		all the players whose names contain that string

###############################################################################

-----changes:---------------------
  April-July, 2004
     TeamWhore - created this file and all its contents
		Created a lot of new functions and moved existing code into them, then moved
			some of those functions to this file, for better organization
		Added a lot of these functions as well, to implement new mod features
		Small updates/fixes to many of the functions, mainly for cosmetic value


****************************************************************************************/


#ifdef UBERSOLDAT_OPTIONAL		//almost all these functions are 'optional'
/*///////////////////////////////////////////////////
Cmd_Gib_f
Simply allows clients to gib themselves via a command.
Formatted so that it can not be used to gib/kill if that
feature is not disabled.
*////////////////////////////////////////////////////
void Cmd_Gib_f(gentity_t *ent)
{
	if (g_allowGib.integer)
	{
		ent->client->ps.persistant[PERS_HWEAPON_USE] = 0;	//copied from Cmd_Kill_f - get off the mg42
		G_Damage(ent, NULL, NULL, NULL, NULL, 1000000, 0, MOD_FALLING);	//kill them

		//TODO: you can perma-gib with this...
	}
}

/*///////////////////////////////////////////////////
Cmd_Clientnum_f
Client command to allow clients to get other players' client
numbers.  The command accepts a name match rather than a full name,
and can display multiple results if there is more than
one match.  For example: "/clientnum bot" would match 'bot', 'robot', 
'lobotomy', '{nOOb}^BoT', 'bother', etc.  Results are printed
to the client who issued the command.
*////////////////////////////////////////////////////
void Cmd_Clientnum_f(gentity_t *ent)
{
	char name[32];
	int nums[MAX_CLIENTS];
	char message[MAX_STRING_CHARS];
	int count = 0;
	int i;

	trap_Argv(1, name, sizeof(name));	//if they didn't enter a name, just return
	if (!name || !*name)
		return;

	if (strlen(name) < 3)	//don't match really short strings
	{
		trap_SendServerCommand(ent-g_entities, va("print \"Enter at least 3 characters to match\n\""));
		return;
	}

	count = ClientNumberFromNameMatch(name, nums);	//get the client numbers that matched

	if (count == 0)	//if there were no matches
	{
		trap_SendServerCommand(ent-g_entities, va("print \"Client not on server\n\""));
		return;
	}

	strcpy(message, va("print \""));	//just the start of a server command

	//add the client name and number for all of the matches, color formatted (and each name/number set is on a new line)
	for (i = 0; i < count; i++)
		strcat(message, va("%s: %s%i%s\n", g_entities[nums[i]].client->pers.netname, S_COLOR_YELLOW, nums[i], S_COLOR_WHITE));

	strcat(message, "\"");

	trap_SendServerCommand(ent-g_entities, message);	//send the message to the client who typed the command
}

/*////////////////////////////////////////////////////
Cmd_Stats_f
Client command to display individual player stats in the
console.  Stats include kills, deaths, kill-death ratio,
gibs, accuracy detailing shots fired and hits, number of
players revived, number of health and ammo packs distributed,
and number of headshots.  Team kills and gibs are also listed
if applicable.
*/////////////////////////////////////////////////////
void Cmd_Stats_f(gentity_t *ent)
{
	char message[MAX_STRING_CHARS];
	gclient_t *client = ent->client;

	int deaths = client->ps.persistant[PERS_KILLED];
	float killRatio = client->pers.kills;

	int shots = client->pers.acc_shots;
	float acc = 0.0f;

	//avoid divide-by-0, will be equal to kills until player dies
	if (deaths > 0)
		killRatio = (float)client->pers.kills / (float)deaths;
		
	//avoid divide-by-0 error, will be 0.0 until the player hits someone
	if (shots > 0)
		acc = ((float)client->pers.acc_hits / (float)shots) * 100.0f;

	strcpy(message, va("print \"Stats for %s ^7this round: \n", client->pers.netname));

	strcat(message, va("^3Kills: ^7%d", client->pers.kills));
	if (client->pers.teamKills > 0)
		strcat(message, va(" (^1%d^7)", client->pers.teamKills));	//only show team kills if greater than 0
	strcat(message, va(", "));

	strcat(message, va("^3Deaths: ^7%d, ", deaths));
	strcat(message, va("^3Kill Ratio: ^7%2.2f, ", killRatio));

	strcat(message, va("^3Gibs: ^7%d \n", client->pers.gibs));
	if (client->pers.teamGibs > 0)
		strcat(message, va(" (^1%d^7)", client->pers.teamGibs));		//only show team gibs if greater than 0
	strcat(message, va(" \n"));

	strcat(message, va("^3Revived: ^7%d, ", client->pers.revives));
	strcat(message, va("^3Packs Given: ^7%d health, %d ammo \n", client->pers.medPacks, client->pers.ammoPacks));
	strcat(message, va("^3Accuracy: ^7%2.2f (%d/%d), ", acc, client->pers.acc_hits, shots));
	strcat(message, va("^3Headshots: ^7%d \n", client->pers.headshots));

	strcat(message, "\"");

	trap_SendServerCommand(ent-g_entities, message);
}

extern void G_Say(gentity_t *ent, gentity_t *target, int mode, const char *chatText);
/*////////////////////////////////////////////////////
Cmd_Omw_f
Allows players to use the "Affirmative, on my way" sound
as a voice chat.  This chat is subject to all spam and
'ignore' restrictions, and is played only to the player's
team.
*/////////////////////////////////////////////////////
void Cmd_Omw_f(gentity_t *ent)
{
	int i;
	gentity_t *te;
	char index[MAX_OSPATH];

	if (!g_omw.integer)
		return;

	if (ent->ignored)	//don't let ignored people use this
	{
		trap_SendServerCommand(ent-g_entities, "print \"^3You are ignored.  Voice chat cancelled\n\"");
		return;
	}

	//subject this to the same spam protections as other chats
	//code copied from G_Voice
	ent->voiceChatSquelch -= (level.time - ent->voiceChatPreviousTime);
	ent->voiceChatPreviousTime = level.time;

	if (ent->voiceChatSquelch < 0)
		ent->voiceChatSquelch = 0;

	if (ent->voiceChatSquelch >= 30000) {
		trap_SendServerCommand(ent-g_entities, "print \"^1Spam Protection^7: VoiceChat ignored\n\"");
		return;
	}

	if (g_voiceChatsAllowed.integer)
		ent->voiceChatSquelch += (34000 / g_voiceChatsAllowed.integer);
	else
		return;

	G_Say(ent, NULL, SAY_TEAM, "Affirmative, on the way!");	//display the message
		
	//loop through all the clients, and play the sound for those on the same team
	for (i = 0; i < MAX_CLIENTS; i++)
	{
		if (level.clients[i].pers.connected != CON_CONNECTED)
			continue;

		if (ent->client->sess.sessionTeam != level.clients[i].sess.sessionTeam)
			continue;

		te = G_TempEntity(ent->s.pos.trBase, EV_GLOBAL_CLIENT_SOUND);

		if (ent->client->sess.sessionTeam == TEAM_RED)
			Q_strncpyz(index, "sound/multiplayer/axis/g-affirmative_omw.wav", sizeof(index));
		else
			Q_strncpyz(index, "sound/multiplayer/allies/a-affirmative_omw.wav", sizeof(index));

		te->s.eventParm = G_SoundIndex(index);
		te->s.teamNum = ent->s.clientNum;
	}
}

/*///////////////////////////////////////////////////
Cmd_Sticky_f
This function allows players to toggle between using normal
grenades and 'sticky' grenades.  Sticky grenades do not bounce,
but will stick to walls, ceilings, objects etc (but not people).
Once enabled, sticky grenades are enabled for that player until
disabled, the player dies, or the level ends.
*////////////////////////////////////////////////////
void Cmd_Sticky_f(gentity_t *ent)
{
	gclient_t *client = ent->client;
	char message[64] = "Sticky grenades ";

	if (!g_stickyNades.integer)
	{
		trap_SendServerCommand(ent-g_entities, va("print \"Sticky grenades are disabled on this server.\n\""));
		return;
	}

	client->ps.selectedSticky = !client->ps.selectedSticky;		//toggle their selection

	if (client->ps.selectedSticky)
		strcat(message, "^1enabled");
	else
		strcat(message, "disabled");

	trap_SendServerCommand(ent-g_entities, va("cp \"%s\n\" 1", message));	//send them the message
}

/*////////////////////////////////////////////////////
Cmd_Smoke_f
This function allows players to toggle between using normal
airstrike canisters and smoke grenades.  Smoke grenades
require only 1/4 of a LT's charge bar (as opposed to 1/2),
and will emit black and coloured smoke for g_smokeGrenades
seconds.  Unlike sticky grenades, these must be enabled for
each use.
*/////////////////////////////////////////////////////
void Cmd_Smoke_f(gentity_t *ent)
{
	gclient_t *client = ent->client;
	char message[64] = "Smoke grenade ";

	if (!g_smokeGrenades.integer)	//make sure they are enabled
	{
		trap_SendServerCommand(ent-g_entities, va("print \"Smoke grenades are disabled on this server.\n\""));
		return;
	}

	if ((g_smokeGrenadesLmt.integer) && (ent->thrownSmoke >= g_smokeGrenadesLmt.integer))
	{
		trap_SendServerCommand(ent-g_entities, va("cp \"Maxmimum number of smoke grenades used (%d)\n\" 1", g_smokeGrenadesLmt.integer));
		return;
	}

	client->ps.selectedSmoke = !client->ps.selectedSmoke;	//set the new value

	//set up the correct message depending on the client's choice
	if (client->ps.selectedSmoke)
		strcat(message, "^1enabled");
	else
		strcat(message, "disabled");

	trap_SendServerCommand(ent-g_entities, va("cp \"%s\n\" 1", message));	//send the message to the client
}

/*//////////////////////////////////////////////////////
Cmd_Playdead_f
This command allows a player to 'play dead', in that
the player will collapse as if he died, and will continue
to lay on the ground as if wounded (animations are consistent
with this as well).  As a clue to the other players though,
players will still collide with his body, unlike with an
actual dead/wounded player.  To return to normal, the command
is issued again, though the player must remain on the ground for
a short time before getting up, and will be unable to move for
another few seconds after rising.
*///////////////////////////////////////////////////////
void Cmd_Playdead_f(gentity_t *ent)
{
	if (!g_playDead.integer)				//just return if it's not enabled
		return;

	if (ent->client->ps.grenadeTimeLeft)	//can't playdead while holding a ticking grenade
		return;

	if (ent->playdeadState == 0)			//if they do it from standing
	{
		//play the death animation
		ent->client->ps.pm_time = BG_AnimScriptEvent(&ent->client->ps, ANIM_ET_DEATH, qfalse, qtrue);
		ent->client->ps.pm_time += 2500;	//make them lay there for at least a short time
		ent->playdeadState = 1;				//set the intermediate state
	}
	else if (ent->playdeadState == 2)		//if they do it from already playing dead
	{
		//play the jump up animation (does this work?)
		ent->client->ps.pm_time = BG_AnimScriptEvent(&ent->client->ps, ANIM_ET_JUMP, qfalse, qtrue);
		ent->client->ps.pm_time += 3000;	//freeze them for a bit after getting up
		ent->playdeadState = 3;				//set the intermediate state
	}
}

extern char* ConcatArgs(int start);
/*////////////////////////////////////////////////////////
Cmd_Pmsg_f
This command allows one player to send a 'private' message (text only)
to another player or players.  The message is sent to all players whose
names contain the match string provided by the sender.  This command is
not subject to spam rules governing text-only messages, but cannot be
used by people who are ignored.
*/////////////////////////////////////////////////////////
void Cmd_Pmsg_f(gentity_t *ent)
{
	char cmd[MAX_TOKEN_CHARS];	

	char name[MAX_STRING_CHARS];
	char nameList[MAX_STRING_CHARS];
	char *msg;
	int matchList[MAX_CLIENTS];
	int count, i;

	if (!g_privateMessages.integer)		//just return if this isn't enabled
		return;

	if (trap_Argc() < 3)
	{
		trap_Argv(0, cmd, sizeof(cmd));	//just for consistency, display the command they entered

		trap_SendServerCommand(ent-g_entities, va("print \"Usage:  %s <match> <message>\n\"", cmd));
		return;
	}

	if (ent->ignored)	//ignored people can't do ANY kind of talking
	{
		trap_SendServerCommand(ent-g_entities, va("print \"^3You are ignored.  Chat cancelled\n\""));
		return;
	}

	trap_Argv(1, name, sizeof(name));
	if (strlen(name) < g_minMsgChars.integer)	//make sure they entered a long enough name match
	{
		trap_SendServerCommand(ent-g_entities, va("print \"You must match at least %i characters of the name\n\"", g_minMsgChars.integer));
		return;
	}

	count = ClientNumberFromNameMatch(name, matchList);

	if (count == 0)
	{
		trap_SendServerCommand(ent-g_entities, va("print \"No matching clients found\n\""));
		return;
	}

	msg = ConcatArgs(2);	//get the whole message
		
	//loop through all the matching clients and display the message for them
	for (i = 0; i < count; i++)
	{
		strcat(nameList, g_entities[matchList[i]].client->pers.netname);	//add their name to a list
		if (i != (count-1))
			strcat(nameList, "^7, ");

		//if the message is short enough to fit in the talk space of the window, display it there
		if ((strlen(msg) + strlen(va("Message from %s^7: ^3", ent->client->pers.netname))) < MAX_SAY_TEXT)
			trap_SendServerCommand(matchList[i], va("chat \"Message from %s^7: ^3%s\" 0", ent->client->pers.netname, msg));
		else	//otherwise, print it in the console, and notify the recipient(s) that they received a message there
		{
			trap_SendServerCommand(matchList[i], va("print \"Message from %s^7: ^3%s\n\"", ent->client->pers.netname, msg));
			trap_SendServerCommand(matchList[i], va("print \"Message from %s^7: ^3Check your console\n\"", ent->client->pers.netname));
		}
	}

	//let the sender know who his message went to
	trap_SendServerCommand(ent-g_entities, va("print \"Message to %s: %s\n\"", nameList, msg));
}

/*********************************** END g_cmds.c functions *************************************/


/*////////////////////////////////////////////////////
Cmd_Push_f
Implements player pushing (with +salute), so that players
can push other players out of their way, or possible hurt/kill
them if g_shoveAmount is set high.
*/////////////////////////////////////////////////////
void Cmd_Push_f(gentity_t* ent)
{
	gentity_t *target;
	trace_t tr;
	vec3_t start, end, forward;
	float shoveAmount;

	if (!g_shove.integer)	//if shoving is disabled, just return
		return;

	if (ent->client->ps.stats[STAT_HEALTH] <= 0)	//dead people can't shove
		return;

	if (level.time < (ent->lastPushTime + 600))		//return if it's too soon since our last shove
		return;

	AngleVectors(ent->client->ps.viewangles, forward, NULL, NULL);

	VectorCopy(ent->s.pos.trBase, start);	//set 'start' to the player's position (plus the viewheight)
	start[2] += ent->client->ps.viewheight;
	VectorMA (start, 128, forward, end);	//put 'end' 128 units forward of 'start'

	//see if we hit anything between 'start' and 'end'
	trap_Trace (&tr, start, NULL, NULL, end, ent->s.number, CONTENTS_BODY);

	//if we didn't hit a player, return
	if (tr.entityNum >= MAX_CLIENTS)
		return;

	target = &(g_entities[tr.entityNum]);

	if ((!target->inuse) || (!target->client))	//if the player is lagged/disconnected/etc
		return;

	if (target->client->ps.stats[STAT_HEALTH] <= 0)	//if they're dead
		return;

	//push them by adding 512*g_shoveAmount units to their velocity, in the direction the pusher is facing
	shoveAmount = 512 * g_shoveAmount.value;
	VectorMA(target->client->ps.velocity, shoveAmount, forward, target->client->ps.velocity);

	ent->lastPushTime = level.time;		//set this so we can't push too often
}

/*///////////////////////////////////////////////////////
Cmd_ThrowKnife_f
Implements knife throwing when a player presses the
drop weapon key while holding a knife.  Players can only
throw a knife every 0.75 seconds.
*////////////////////////////////////////////////////////
void Cmd_ThrowKnife_f(gentity_t *ent)
{
	vec3_t velocity, offset, org, mins, maxs;
	trace_t tr;
	gentity_t *ent2;
	gitem_t *item = BG_FindItemForWeapon(WP_KNIFE);

	if (g_throwableKnives.integer == 0)
		return;

	if (level.time < (ent->thrownKnifeTime + 750))		//can throw one every 0.75 seconds
		return;

	//return if we have no more knives to throw (-1 is infinite)
	if ((ent->client->ps.ammo[BG_FindAmmoForWeapon(WP_KNIFE)] == 0) && (g_throwableKnives.integer != -1))
		return;

	//some of this is copied from the drop weapon code in g_active.c
	AngleVectors(ent->client->ps.viewangles, velocity, NULL, NULL);
	VectorScale(velocity, 64, offset);
	offset[2] += ent->client->ps.viewheight/2;
	VectorScale(velocity, 800, velocity);		//make the knife go pretty quickly
	velocity[2] += 50 + random() * 35;
		
	VectorAdd(ent->client->ps.origin, offset, org);
	
	VectorSet(mins, -ITEM_RADIUS, -ITEM_RADIUS, 0);
	VectorSet(maxs, ITEM_RADIUS, ITEM_RADIUS, 2*ITEM_RADIUS);	//these have something to do with making the item sit on the ground

	trap_Trace (&tr, ent->client->ps.origin, mins, maxs, org, ent->s.number, MASK_SOLID);
	VectorCopy(tr.endpos, org);
	
	G_Sound(ent, G_SoundIndex("sound/weapons/knife/knife_slash1.wav"));		//play a sound
	ent2 = LaunchItem(item, org, velocity, ent->client->ps.clientNum);		//actually throw the knife

	ent2->touch = Touch_Knife;	//set this to our own handler so we can customize what happens
	ent2->parent = ent;			//so we know who threw the knife

	if (g_throwableKnives.integer > 0)
		ent->client->ps.ammo[BG_FindAmmoForWeapon(WP_KNIFE)]--;		//get rid of a knife

	ent->thrownKnifeTime = level.time;							//mark this so they can't throw a ton of them in a row
}

/*////////////////////////////////////////////
Touch_Knife
Called whenever a client touches a thrown knife.
Made this a separate function rather than edit 4 or 5 id funcs.
Will either hurt the client or have them pick up the knife.
Most of the pickup code is copied from Touch_Item.
*/////////////////////////////////////////////
void Touch_Knife(gentity_t *ent, gentity_t *other, trace_t *trace)
{
	qboolean hurt = qfalse;
	ent->active = qfalse;	//set this just to be safe, but shouldn't be necessary

	//make sure this is a person, and that they're not dead
	if (!other->client)
		return;
	if (other->health < 1)
		return;		// dead people can't pickup

	if (VectorLength(ent->s.pos.trDelta) != 0)	//if the knife is still moving (not on ground yet)
	{	
		//if friendly fire is off, or if the player who touched the thrown knife is on the other team
		if ((g_friendlyFire.integer) || (!OnSameTeam(other, ent->parent)))
		{
			int i;
			int sound;
			int damage = g_knifeDamage.integer;		//set the damage to its max value
			damage -= rand() % 10;					//subtract 0-9 to make it a little varied

			if (damage <= 0)		//make sure there's still some damage
				damage = 1;

			//pick a random sound to play
			i = rand() % 3;
			if (i == 0)
				sound = G_SoundIndex("/sound/weapons/knife/knife_hit1.wav");
			else if (i == 1)
				sound = G_SoundIndex("/sound/weapons/knife/knife_hit2.wav");
			else
				sound = G_SoundIndex("/sound/weapons/knife/knife_hit3.wav");

			G_Sound(other, sound);	//play the sound
			G_Damage(other, ent->parent, ent->parent, NULL, trace->endpos, damage, 0, MOD_KNIFE_THROWN);	//hurt them! (for now, no dir, maybe change later)
			hurt = qtrue;	//mark that this knife damaged someone
		}
	}

	if (hurt == qfalse)		//if we didn't hurt anyone (can't do 'else' because of nested if statements above)
	{
		int	makenoise = EV_ITEM_PICKUP;

		G_LogPrintf("Item: %i %s\n", other->s.number, ent->item->classname);	//log it

		if (g_throwableKnives.integer > 0)
			other->client->ps.ammo[BG_FindAmmoForWeapon(WP_KNIFE)]++;	//add the knife

		//don't let us get more knives than the max
		if (g_maxKnives.integer != 0)	//max of 0 means no max
		{
			if (other->client->ps.ammo[BG_FindAmmoForWeapon(WP_KNIFE)] > g_maxKnives.integer)
				other->client->ps.ammo[BG_FindAmmoForWeapon(WP_KNIFE)] = g_maxKnives.integer;
		}

		//no idea, copied from Touch_Item
		if (ent->noise_index)	
		{
			makenoise = EV_ITEM_PICKUP_QUIET;
			G_AddEvent(other, EV_GENERAL_SOUND, ent->noise_index);
		}

		//also from Touch_Item
		//send the pickup event
		if (other->client->pers.predictItemPickup)
			G_AddPredictableEvent(other, makenoise, ent->s.modelindex);
		else
			G_AddEvent(other, makenoise, ent->s.modelindex);
	}

	ent->freeAfterEvent = qtrue;	//get rid of this item
	ent->flags |= FL_NODRAW;
	ent->r.svFlags |= SVF_NOCLIENT;
	ent->s.eFlags |= EF_NODRAW;
	ent->r.contents = 0;
	ent->nextthink = 0;
	ent->think = 0;
}

/*///////////////////////////////////////////////
Cmd_KickGrenade_f
Implements grenade kicking whenever a player presses +salute
while standing near a grenade (with some restrictions).
This is not an extremely cheap function to call, but it
shouldn't create any problems in virtually any situation.
There have been issues with it causing rare game freezes
while it was being developed, but these no longer seem to occur. 
*////////////////////////////////////////////////
void Cmd_KickGrenade_f(gentity_t *ent)
{
	gentity_t *e, *grenade = NULL;
	float dist, bestdist = -1;

	if (!g_kickableNades.integer)	//if it's not enabled, skip everything and just return
		return;

	//loop through every non-player entity, and find the closest valid grenade
	//ordered for minimal calculations and impact - SHOULD be unnoticable
	for (e = g_entities + MAX_CLIENTS; e < &g_entities[level.num_entities]; e++)
	{
		//if it's invalid or unused, skip it
		if (!e || !e->inuse)
			continue;

		//if it's not a grenade, skip it (have to check both, flamebarrels etc use same mod, and airstrikes are 'grenades')
		if ((e->methodOfDeath != MOD_GRENADE) || (Q_stricmp(e->classname, "grenade") != 0))
			continue;

		//can't kick sticky grenades
		if (e->s.eFlags & EF_STICK)
			continue;

		//cvar must be ANDed with 1 to allow players to kick their own grenades
		if ((e->parent == ent) && (!(g_kickableNades.integer & 1)))
			continue;

		//cvar must be ANDed with 2 to allow players on the grenade owner's team to kick it
		if ((!(g_kickableNades.integer & 2)) && (e->parent != ent) && (OnSameTeam(ent, e->parent)))
			continue;

		//cvar must be ANDed with 4 to allow players on the other team to kick the grenade
		if ((!(g_kickableNades.integer & 4)) && (!OnSameTeam(ent, e->parent)))
			continue;

		//can't kick grenades that are above waist-height
		if (e->r.currentOrigin[2] >= ent->r.currentOrigin[2])
			continue;

		//this doesn't work (probably would have to change the dotConst again
		//seemingly unnecessary, since distance must be so low anyway, so remove it to save on calculations
		/*if (!infront(ent, e))
			continue;*/

		//find the nearest valid grenade
		dist = VectorDistance(ent->r.currentOrigin, e->r.currentOrigin);
		if ((dist < bestdist) || (bestdist == -1))
		{
			grenade = e;
			bestdist = dist;
		}
	}

	if ((grenade) && (bestdist <= 70))	//if we found a grenade and it's close enough to kick
	{
		//don't kick it if it's moving too fast or about to explode(causes errors? SOMEthing does...)
		if ((grenade->s.pos.trDelta[2] <= 160) && (VectorLength(grenade->s.pos.trDelta) <= 180) && (grenade->nextthink - level.time >= 400))
		{
			float speed = 800.0f;		//default value, based on some testing
			vec3_t angles, forward;

			VectorClear(angles);		//only angles[1] is used
			//set it to the direction we're facing, +/- 15 degrees
			angles[1] = ent->client->ps.viewangles[1] + (crandom() * 15.0f);

			//get the 'forward' vector based on the direction we're facing
			AngleVectors(angles, forward, NULL, NULL);
			//add some randomness to the 'height' of the kick
			forward[2] = 0.5f + (crandom() * 0.15f);
			VectorNormalizeFast(forward);				//re-normalize the vector

			speed += (crandom() * 150.0f);				//add some randomness to the speed/distance of the kick
			VectorScale(forward, speed, forward);		//and multiply the vector by the speed

			//play the grenade throwing sound, just to let them know they successfully kicked it
			G_Sound(ent, G_SoundIndex("sound/weapons/grenade/grenlf1a.wav"));

			trap_UnlinkEntity(grenade);					//probably not necessary, may prevent an occasional error
			grenade->s.pos.trType = TR_GRAVITY;			//have to set this in case it was reset to TR_STATIONARY
			grenade->s.pos.trTime = level.time - 50;	//move a little on the first frame

			//set the grenade's initial position
			VectorCopy(grenade->r.currentOrigin, grenade->s.pos.trBase);
			VectorCopy(forward, grenade->s.pos.trDelta);	//essentially, set the velocity
			SnapVector(grenade->s.pos.trDelta);				//"save network bandwith" apparently...rounds everything to integers
			trap_LinkEntity(grenade);						//re-link the grenade...have to if we unlinked it
		}
	}
}

/*///////////////////////////////////////////////
Cmd_KickAS_f
Similar to Cmd_KickGrenade_f, this function implements
kicking of airstrike canisters with the +salute command.
However, these cannot be kicked nearly as far as grenades,
and only the LT who 'owns' the canister can kick it.
Enough was different to warrant its own function.
*////////////////////////////////////////////////
void Cmd_KickAS_f(gentity_t *ent)
{
	gentity_t *e, *grenade=NULL;
	float dist, bestdist=-1;

	if (!g_kickableAS.integer)	//just return if this isn't enabled
		return;

	if (ent->client->ps.stats[STAT_PLAYER_CLASS] != PC_LT)	//non lt's can't have any airstrikes to kick...
		return;

	//loop through every non-player entity, and find the nearest valid airstrike canister (if any)
	for (e = g_entities + MAX_CLIENTS; e < &g_entities[level.num_entities]; e++)
	{
		if (!e || !e->inuse)	//skip unused entities
			continue;

		//skip non-airstrike-canisters
		if ((Q_stricmp(e->classname, "grenade") != 0) || (e->s.weapon != WP_SMOKE_GRENADE))
			continue;

		//only the lt who threw this airstrike can kick it
		if (e->parent != ent)
			continue;

		//it must be at or below waist-height
		if (e->r.currentOrigin[2] >= ent->r.currentOrigin[2])
			continue;

		//it can't be moving/bouncing (unlike grenades)
		if (VectorLength(e->s.pos.trDelta) > 0)
			continue;

		//find the closest matching airstrike canister
		//this probably isn't necessary, since normally a lt can only have one canister out at a time
		//but if the lt charge time was VERY low, it might be possible to have multiple ones
		dist = VectorDistance(ent->r.currentOrigin, e->r.currentOrigin);
		if ((dist < bestdist) || (bestdist == -1))
		{
			grenade = e;
			bestdist = dist;
		}
	}

	if ((grenade) && (bestdist <= 70))	//if we found a canister that is close enough to kick
	{
		float speed = 500.0f;	//much lower speed for these compared to grenades
		vec3_t angles, forward;

		//virtually the same as in Cmd_KickGrenade_f
		VectorClear(angles);
		angles[1] = ent->client->ps.viewangles[1] + (crandom() * 15.0f);

		AngleVectors(angles, forward, NULL, NULL);
		forward[2] = 0.4f + (crandom() * 0.15f);	//slightly lower arc
		VectorNormalizeFast(forward);

		speed += (crandom() * 150.0f);
		VectorScale(forward, speed, forward);

		trap_UnlinkEntity(grenade);
		grenade->s.pos.trType = TR_GRAVITY;
		grenade->s.pos.trTime = level.time - 50;
		
		VectorCopy(grenade->r.currentOrigin, grenade->s.pos.trBase);
		VectorCopy(forward, grenade->s.pos.trDelta);
		SnapVector(grenade->s.pos.trDelta);
		trap_LinkEntity(grenade);
	}
}

/********************************** END g_active.c functions *************************************/


extern void G_ExplodeMissile(gentity_t *ent);	//used in below function
/*///////////////////////////////////////////////
weapon_smokeGrenade
The 'think' function for smoke grenades - makes them
smoke and plays the hissing sound.
This should never explicitly be called by anything.
*////////////////////////////////////////////////
void weapon_smokeGrenade(gentity_t *ent)
{
	ent->s.eFlags |= EF_SMOKINGBLACK;		//make it emit thick black smoke
	ent->s.loopSound = G_SoundIndex("sound/world/steam.wav");	//play the hissing steam sound as long as the can exists

	ent->nextthink = level.time + ((g_smokeGrenades.integer - 1) * 1000);	//do this until g_smokeGrenades seconds is up
	ent->think = G_ExplodeMissile;			//make it explode and get rid of it
}

/********************************** END 'think' functions ****************************************/


/*/////////////////////////////////////////////////
PlayerNeedsAmmo
Maleficus' function to check whether a player currrently
has any use for an ammo pack.  This allows us to prevent
the player from picking up the pack if he does not need it.
TW - modified to include new cvars and to return quicker
*//////////////////////////////////////////////////
qboolean PlayerNeedsAmmo(gentity_t *ent)
{
	int nades, weapon, i;
	int numClips;					//BOTS TW - since g_maxClips overrides the default

	if (!g_noAmmoWhore.integer)		//BOTS TW - added cvar check - if this feature isn't enabled, let them pick up the pack
		return qtrue;

	//check nades first!
	switch (ent->client->ps.stats[STAT_PLAYER_CLASS])
	{
	case PC_ENGINEER:
		nades = 8;
		break;
	case PC_SOLDIER:
		nades = 4;
		break;
	case PC_LT:
		nades = g_LTNades.integer;		//BOTS TW - changed to support new cvar
		break;
	case PC_MEDIC:
		nades = g_medicNades.integer;	//BOTS TW - changed to support new cvar
		break;
	default:
		nades = 1;
		break;
	}

	//if short on nades, pickup!
	if (ent->client->sess.sessionTeam == TEAM_RED)
		weapon = WP_GRENADE_LAUNCHER;
	else
		weapon = WP_GRENADE_PINEAPPLE;

	if (ent->client->ps.ammoclip[BG_FindClipForWeapon(weapon)] < nades)
		return qtrue;

	//if a med, and short on needles, pickup!
	if (g_ammoGivesSyringe.integer)				//BOTS TW - added support for cvar enabling/disabling of this feature
	{	
		if (ent->client->ps.stats[STAT_PLAYER_CLASS] == PC_MEDIC)
		{
			if (ent->client->ps.ammoclip[BG_FindClipForWeapon(WP_MEDIC_SYRINGE)] < 10)
				return qtrue;
		}
	}

	//BOTS TW - added check for throwing knives, if the cvar is enabled
	//if short on knives, he needs ammo
	if (g_throwableKnives.integer && g_ammoGivesKnife.integer)
	{
		if ((ent->client->ps.ammoclip[BG_FindClipForWeapon(WP_KNIFE)] < g_maxKnives.integer) || (g_maxKnives.integer == 0))
			return qtrue;
	}

	//if need a helmet, pickup!
	if (g_ammoGivesHelmet.integer)				//BOTS TW - added support for cvar enabling/disabling of this feature
	{
		if (ent->client->ps.eFlags & EF_HEADSHOT)
			return qtrue;
	}

	//check pistols
	if (ent->client->sess.sessionTeam == TEAM_RED)
		weapon = WP_LUGER;
	else
		weapon = WP_COLT;

	//if need pistol ammo, pickup!
	if (ent->client->ps.ammo[BG_FindAmmoForWeapon(weapon)] < ammoTable[weapon].maxclip*4)
		return qtrue;

	//loop through remaining weapons and if need ammo, pickup!
	for (i = 0; i < MAX_WEAPS_IN_BANK_MP; i++)
	{
		weapon = weapBanksMultiPlayer[3][i];

		if (COM_BitCheck(ent->client->ps.weapons, weapon))	//BOTS TW - if we have this weapon
		{
			if (weapon == WP_FLAMETHROWER)
			{
				//BOTS TW - flamers only have one 'clip'
				if (ent->client->ps.ammoclip[BG_FindAmmoForWeapon(weapon)] < ammoTable[weapon].maxclip)
					return qtrue;
			}
			
			//BOTS TW - other guns currently support at most 3 extra clips (or g_maxClips)
			if (g_maxClips.integer)
				numClips = g_maxClips.integer;
			else
				numClips = 3;

			if (ent->client->ps.ammo[BG_FindAmmoForWeapon(weapon)] < ammoTable[weapon].maxclip*numClips)
				return qtrue;
		}
	}

	return qfalse;		//player doesn't need any ammo, so don't let him pick up the pack!
}
#endif

/*///////////////////////////////////////////////
ClientNumberFromNameMatch
This function does what the name says - it takes
a string representing a partial name, and returns the
client numbers of all the players whose names contain
that string.  The client numbers are stored in 'matches'
and the number of matches is returned.  This function does
all the string handling you should need, so it is safe to pass
unformatted names to it
*////////////////////////////////////////////////
int ClientNumberFromNameMatch(char *name, int *matches)
{
	int i, textLen;
	char nm[32];
	char c;
	int index = 0;

	Q_strncpyz(nm, name, sizeof(nm));	//copy the name match
	Q_CleanStr(nm);						//get rid of any color codes etc
	textLen = strlen(nm);
	c = *nm;							//the first letter of the name match

	for (i = 0; i < level.maxclients; i++)	//loop through all the possible clients
	{
		int j, len;
		char playerName[32];

		if ((!g_entities[i].client) || (g_entities[i].client->pers.connected != CON_CONNECTED)) //if this client isn't connected, skip them
			continue;
		
		Q_strncpyz(playerName, g_entities[i].client->pers.netname, sizeof(playerName));	//copy the player's name
		Q_CleanStr(playerName);		//remove any color codes etc
		len = strlen(playerName);

		for (j = 0; j < len; j++)	//go through the player's name letter by letter
		{
			if (tolower(c) == tolower(playerName[j]))	//compare the first letter of the name match to each letter in the player's name
			{
				//if they're the same, check to see if the whole name match can be found
				if (!Q_stricmpn(nm, playerName+j, textLen))	
				{
					matches[index] = i;		//we have a match, so record the client number
					index++;				//increment the index counter
					break;					//break out of the 'j' FOR loop because we already matched this client
				}
			}
		}
	}

	return index;	//return the number of matched client names
}

